package POE::Component::Generic;

use POE 0.31;
use POE::Wheel::Run;
use POE::Filter::Line;
use POE::Filter::Reference;
use Carp qw(carp croak);
use Devel::Symdump;
use vars qw($AUTOLOAD);
use Config;

use strict;

our $VERSION = '0.01';

sub AUTOLOAD {
    my $self = shift;
    my $method = $AUTOLOAD;
    $method =~ s/.*:://;
    return unless $method =~ /[^A-Z]/;
    
    warn "autoload method $method" if ($self->{debug});
    
    my $hash = shift;
    croak "first param to $method must be a hash ref of options" 
              unless ref($hash) eq 'HASH';
    
    $hash->{wantarray} = wantarray() unless (defined($hash->{wantarray}));
    
    # use ->call() so that $generic->method() happens in order
    $poe_kernel->call( $self->session_id() => $method => $hash => @_ );
}

##########################################################################
sub spawn {
    goto &new;
}

##################################################
sub new {
    my $package = shift;
    croak "$package needs an even number of parameters" if @_ & 1;
    my %params;
    {
        my %p = @_;
        while( my( $k, $v ) = each %p ) {
            $params{ lc $k } = $v;
        }
    }

    my $options = delete ( $params{'options'} );
    unless( $params{package} ) {
        croak "Please specify a package";
    }

    # map of commands to packages
    $params{cmd_map} = {};
    # param storage
    $params{store} = {};

    my $self = bless(\%params, $package);

    POE::Component::Generic::Child::package_load( $self->{package} );
    $self->package_map;

    $self->{session_id} = POE::Session->create(
        object_states => [
            $self => { (map { $_ => 'request' } keys %{$self->{cmd_map}}) },
            $self => [ qw(_start shutdown _child
                          wheel_close wheel_err wheel_out wheel_stderr) ],
        ],
        ( ( defined ( $options ) and ref ( $options ) eq 'HASH' ) ? ( options => $options ) : () ),
    )->ID();

    warn "session $self->{session_id} created for $params{package}" 
          if ($self->{debug});
    
    return $self;
}

##################################################
# Build a map of all methods => package
sub package_map
{
    my( $self ) = @_;
    my @methods = $self->package_methods( $self->{package} );
    
    my %OK;
    if( $self->{methods} ) {
        @OK{ @{ $self->{methods} } } = (1) x @{ $self->{methods} };
    } 
    
    foreach my $p ( @methods ) {
        my ($pk,$sub) = ($p =~ m/^(.+)\:\:([^\:]+)/);
    
        next if $self->{cmd_map}->{ $sub };
        next if $self->{methods} and not $OK{ $sub };
        next unless ($sub =~ /[^A-Z_0-9]$/);
        next if ($sub =~ m/^_/ || $sub =~ m/(carp|croak|confess)$/);
        my $o = $p;
        
        if (defined &$o) {
            $self->{cmd_map}->{$sub} = $pk;
        }
    }
    

}

##################################################
# Get a list of all methods from the package
sub package_methods
{
    my( $self, $package ) = @_;
    my $isa = Symbol::qualify_to_ref( "ISA", $package );
    my @obj = Devel::Symdump->functions( $package );
    foreach my $subpack ( @{ *$isa } ) {
        next if $subpack eq 'Exporter';
        push @obj, $self->package_methods( $subpack );
    }
    # we can't distinguish methods from functions :-/
    return @obj
}    






##########################################################################
# POE related object methods

sub _start {
    my ($kernel,$self) = @_[KERNEL,OBJECT];

    if ( $self->{alias} ) {
        $self->{name} = $self->{alias};
        $kernel->alias_set( $self->{alias} );
        $self->{debug} and warn "alias is $self->{alias}";
    } 
    else {
        $self->{name} = "poe-generic";
        $kernel->refcount_increment( $self->{session_id} => __PACKAGE__ );
    }
    
    $poe_kernel->sig( CHLD => '_child' );
    my %prog = (Program => \&process_requests);
    if ($self->{alt_fork}) {
    
        my $perl = $^X;
        if( $ENV{HARNESS_PERL_SWITCHES} ) {
            $perl .= " $ENV{HARNESS_PERL_SWITCHES}";
        }
                        
        %prog = (Program =>  "$perl -M".__PACKAGE__
                  ." -I".join(' -I',@INC)
                  ." -e '".__PACKAGE__."::process_requests(1)'");
        $self->{debug} and 
            warn "Launching $prog{Program}";
    }
    
    $self->{wheel} = POE::Wheel::Run->new(
        %prog,
        CloseOnCall  => 0,
        StdinFilter  => POE::Filter::Reference->new(),
        StdoutFilter => POE::Filter::Reference->new(),
        StderrFilter => POE::Filter::Line->new(),
        StdoutEvent  => 'wheel_out',
        StderrEvent  => 'wheel_stderr',
        ErrorEvent   => 'wheel_err',
        CloseEvent   => 'wheel_close',
    );

    # Tell the other side to create an object
    $self->{object_options} ||= [];
    unless( ref $self->{object_options} ) {
        $self->{object_options} = [ $self->{object_options} ] 
    }

    my $new = {  req     => 'setup',
                 debug   => $self->{debug},
                 args    => $self->{object_options},
                 package => $self->{package},
                 name	 => $self->{name}
              };
    $new->{size} = $self->{size} if $self->{size};
              
    $self->{debug} and warn "Ask to create object";
    $self->{wheel}->put( $new );

    undef;
}

######################################################
# Send request to child process
sub request 
{
    my ( $self,$state ) = @_[OBJECT,STATE];
    my $sender = $_[SENDER]->ID;
  
    warn "$$: processing request $state\n" if ($self->{debug});
    
    # Get the arguments
    my $args;
    if (ref($_[ARG0]) eq 'HASH') {
        $args = { %{ $_[ARG0] } };
    } 
    else {
        warn "first parameter must be a ref hash, trying to adjust. "
            ."(fix this to get rid of this message)";
        $args = { @_[ARG0 .. $#_ ] };
    }
 
    if ($self->{wheel}) {
        # If we have an {event}, it means the user wants *something* back
        if( $args->{event} and not defined $args->{wantarray} ) {
            $args->{wantarray} = 0;
        }
        my $params = {
            state     => $state,
            event     => $args->{event},
            wantarray => $args->{wantarray},
            session   => ($args->{session}||$sender),
            args      => [ @_[ ARG1 .. $#_ ] ],
        };
        
        # param storage
        if (keys %$args) {
            my $id = $self->{id}++;
            # id to match in param storage
            $params->{id} = $id;
            $self->{store}->{$id} = $args;
        }

        # if we have an event to report to...make sure it stays around
        if ($args->{event}) {
            $poe_kernel->refcount_increment( $sender => $self->{name} );
        }
        $self->{debug} and warn "request put";
        $self->{wheel}->put($params);
    }
    elsif( $self->{debug} ) {
        warn "No wheel";
    }
    
    undef;
}

######################################################
# Child process sent us a response
sub wheel_out 
{
    my ($self,$input) = @_[ OBJECT,ARG0 ];

    $self->{debug} and warn "wheel_out";
    if( $input->{response} ) {
        if( $input->{response} eq 'new' ) {
            $self->{child_PID} = $input->{PID};
            $self->{debug} and warn "Child PID=$input->{PID}";
        }
        else {
            warn "Unknown child response $input->{response}";
        }
        return;
    }

    if (defined $input->{id}) {
        my $id = delete $input->{id};
        # splice in stored data, because we might not trust other side
        @{ $input }{ keys %{$self->{store}->{$id}} }
              = values %{$self->{store}->{$id}};
        delete $self->{store}->{$id};
    }

    my $session = delete $input->{session};
    my $event = delete $input->{event};

    if ($event) {
        my $res = $input->{result} || [];
        $self->{debug} and warn "Reply to $session/$event";
        $poe_kernel->post( $session => $event, $input => @$res );
        $poe_kernel->refcount_decrement( $session => $self->{name} );
    }
  
    undef;
}

sub wheel_stderr {
    my ($kernel,$self,$input) = @_[KERNEL,OBJECT,ARG0];

    warn "ERR:$self->{name}: $input\n" if ($self->{debug});
}

sub wheel_err {
    my ($self, $operation, $errnum, $errstr, $wheel_id) = @_[OBJECT, ARG0..ARG3];
    
    warn "Wheel:$self->{name}: Wheel $wheel_id generated $operation error $errnum: $errstr\n" 
            if ($self->{debug});
}

sub wheel_close {
    my $self = $_[OBJECT];
    
    warn "Wheel closed\n" if ($self->{debug});
    
#   warn "$self->{package} Wheel closed, ieeeeeeee!\n";
}

sub _child
{
    my( $self, $name, $PID, $ret ) = @_[ OBJECT, ARG0..ARG2 ];
    unless( $PID == $self->{child_PID} ) {
        $self->{debug} and warn "Got CHLD for $PID";
        return;
    }
    $self->{debug} and warn "Child $PID exited with $ret";
    $poe_kernel->sig_handled;
    return;
}

############################################################################
# Dual event and object methods

sub shutdown {
    unless (UNIVERSAL::isa($_[KERNEL],'POE::Kernel')) {
        if ($poe_kernel) {
            $poe_kernel->call(shift->session_id() => 'shutdown' => @_);
        }
        return;
    }
    
    my ($kernel,$self) = @_[KERNEL,OBJECT];

    # remove alias or decrease ref count
    if ($self->{alias}) {
        $kernel->alias_remove($_) for $kernel->alias_list();
    } else {
        $kernel->refcount_decrement($self->session_id() => __PACKAGE__);
    }
    
    if ($self->{wheel}) {
        $self->{wheel}->shutdown_stdin;
    }
    undef;
}


# Object methods

sub session_id {
    shift->{session_id};
}

sub yield {
    my $self = shift;
    $poe_kernel->post($self->session_id() => @_);
}

sub call {
    my $self = shift;
    $poe_kernel->call($self->session_id() => @_);
}

sub DESTROY {
    if (UNIVERSAL::isa($_[0],__PACKAGE__)) {
        $_[0]->shutdown();
    }
}

##########################################################################
# Main Wheel::Run process sub

sub process_requests {
    my $alt_fork = shift || 0;
    
    binmode(STDIN);
    binmode(STDOUT);
    STDOUT->autoflush(1);

    my $runner = POE::Component::Generic::Child->new( 
                name=>__PACKAGE__,
                size=>4096, 
                debug=>0, 
                proc=>$0, 
                alt_fork=>$alt_fork
            );
    $runner->loop;
}


#######################################
package POE::Component::Generic::Child;
use strict;

sub new
{
    my( $package, %params ) = @_;
    my $self = bless { %params }, $package;
    $self->{filter} = POE::Filter::Reference->new();
    # there's room for other callbacks
    $self->{callbacks} = {};
    return $self;
}

sub loop
{
    my( $self ) = @_;
    
    my $raw;
	
    $self->status( 'startup' );

    READ:
    while ( sysread ( STDIN, $raw, $self->{size} ) ) {
        $self->status( 'request' );
        my $requests = $self->{filter}->get([$raw]);

        ################################
        unless ($self->{obj}) {
            my $req = shift @{$requests};
            unless( ref( $req ) eq 'HASH' and $req->{req} eq 'setup' ) {
                die "First request must be a setup";
            }
            $self->OOB_req( $req );
        }
        
        foreach my $req (@{$requests}) {
            if( $req->{req} ) {
                $self->OOB_req( $req );
            }
            else {
                $self->request( $req );
            }
        }
        $self->status( 'read' );
    }
}

##################################################
sub status
{
    my $self = shift;
    $0 = join ' ', $self->{proc}, $self->{name}, @_;
    
    $self->{debug} and warn join ' ', @_;
}

##################################################
sub reply
{
    my( $self, $resp ) = @_;
    $self->{debug} and warn "reply";
    my $replies = $self->{filter}->put( [ $resp ] );
    print STDOUT @$replies;
}




##################################################
sub request
{
    my( $self, $req ) = @_;
    
    my $func = $req->{state};
    $self->{debug} and warn "method=$func";
    
    my $obj = $self->{obj};

    eval {
        if( $req->{wantarray} ) {
            $req->{result} = [ $obj->$func(@{$req->{args}}) ];
        } 
        elsif( defined $req->{wantarray} ) {
            $req->{result} = [ scalar $obj->$func(@{$req->{args}}) ];
        }
        else {
            $obj->$func( @{ $req->{args} } );
        }
    };

    if ($@) {
        $req->{error} = $@;
        delete $req->{result};
    }

    if( defined $req->{event} ) {
        $self->reply( $req );
    }
    elsif( $req->{error} ) {
        warn $req->{error};
    }

}

##################################################
sub OOB_req
{
    my( $self, $req ) = @_;

    $self->status( 'OOB' );

    my $func = $req->{req};
    if( $func eq 'setup' ) {
        $self->OOB_setup( $req );
    }
    else {
        warn "Unknown OOB request $func";
    }
}

##################################################
sub OOB_setup
{
    my( $self, $req ) = @_;
    
    foreach my $f ( qw( name size debug ) ) {
        next unless exists $req->{$f};
        $self->{$f} = $req->{$f};
        $self->{debug} and warn "Setting $f=$self->{$f}";
    }
    
    $self->{debug} and warn "build object $req->{package}";
    $self->{obj} = object_build( $req->{package}, $req->{args} );

    $self->{debug} and warn "object=$self->{obj}";
    $self->reply( { PID=>$$, response=>'new' } );
}



##################################################
sub OOB_callback
{
    my( $self, $req ) = @_;
    
            ################################
#            if ($func eq 'option_callback') {
#                if (@{$req->{args}}) {
#                    # set the callback event
#                    $callbacks{$func} = $req->{args}->[0];
#                    # then set a coderef to post that back
#                    $t->$func(sub {
#                        # $obj, $option, $is_remote, $is_enabled, $was_enabled, $buf_position
#                        shift; # don't need and can't send the object
#                        $req->{result} = [ @_ ];
#                        $req->{event} = $callbacks{$func};
#                        my $rep = $filter->put( [ $req ] );
#                        print STDOUT @$rep;
#                    });
#                } else {
#                    $req->{result} = $callbacks{$func};
#                    my $rep = $filter->put( [ $req ] );
#                    print STDOUT @$rep;
#                }
#                next;
#            }
}

##################################################
sub object_build
{
    my( $package, $args ) = @_;
    my $ctor = package_load( $package );
    die "Can't find constructor for package $package" unless $ctor;
    return $package->can($ctor)->( $package, @$args );
}


##################################################
# Load the user package
sub package_load
{
    my( $package ) = @_;
    my $ctor = find_ctor( $package );
    return $ctor if $ctor;		# package already loaded
    eval "use $package";
    die $@ if $@;
    return find_ctor( $package );
}

##################################################
sub find_ctor
{
    my( $package ) = @_;
    foreach my $ctor ( qw( new spawn create ) ) {
        return $ctor if $package->can( $ctor );
    }
    return;
}


1;

__END__

=head1 NAME

POE::Component::Generic - A POE component that provides non-blocking access to a blocking object.

=head1 SYNOPSIS

    use POE::Component::Generic;

    my $poco = POE::Component::Generic->new(
        package => 'Net::Telnet',

        # optional; You can use $poco->session_id() instead
        alias => 'telnet',
        # optional; 1 to turn on debugging
        debug => 1,
        # optional; Options passed to the internal session
        options => { trace => 1 },
        # optional; Options passed to Net::Telnet->new()
        object_options => [ ],
    );

    # Start your POE session, then...

    $kernel->post('telnet' => 
          open => { event => 'result' },"rainmaker.wunderground.com");

    sub result {
        my ($kernel, $ref, $result) = @_[KERNEL, ARG0, ARG1];

        if( $ref->{error} ) {
            die join(' ', @{ $ref->{error} ) . "\n";
        }
        print "connected: $result\n";
    }

=head1 DESCRIPTION

POE::Component::Generic is a L<POE> component that provides a non-blocking
wrapper around any object.  It works by creating a L<POE::Wheel::Run> and
creating the object in the child process.  Method requests are then
serialised and sent to the child to be handled.  Return values are posted
back to your session. This means that requests and return values must
survive serialisation.

Method calls are wrapped in C<eval> so that errors may be propagated back to
your session.  See L<OUTPUT>.

Output to STDERR in the child is shown only if C<debug> is set.

Output to STDOUT in the child will MESS THINGS UP.  So don't do that.


=head1 METHODS

=head2 new

Takes a number of arguments, all but C<package> are optional. 

=over 4

=item alias

Session alias to register with the kernel.  Default is none.

=item alt_fork

Set to C<true> if you want to run another perl instance.  That is, the child
process will C<exec> a new instance of C<perl> using C<$^X> to do the work. 
C<@INC> is preserved.  If present, C<$ENV{HARNESS_PERL_SWITCHES}> is used.

Using C<alt_fork> might help save memory; while the child process will only
contain C<POE::Component::Generic> and your object, it will not be able to
share as many memory pages with other processes.

Default is false.

=item debug

Set to C<true> to see component debug information, such as anything
output to STDERR by your code.  Default to C<false>.

=item methods

An array ref containing methods that you want the component to expose. If
not specified, all the methods of C<package> and its super-classes are
exposed.

Note that methods that being with C<_> or don't end with a lower case letter
(C<a-z>) are excluded, as well as methods that end in C<carp>, C<croak> and
C<confess>.

=item options

A hashref of L<POE::Session> options that are passed to the component's
session creator.

=item object_options

An optional array ref of options that will be passed to the object
constructor.

=item package

Package name that is used to create the object.  

Object creation happens in the child process.  The package is loaded, if
it isn't already, then a constructor is called with C<object_options> as
the parameters.  The constructor is a package method named C<new>,
C<spawn> or C<create>, searching in that order.

=back




=head2 session_id

Takes no arguments, returns the L<POE::Session> ID of the component. Useful
if you don't want to use aliases.


=head1 METHOD CALLS

There are 4 ways of calling methods on the object.

All methods need a data hashref that will be handed back to the return
event.  This data hash is discussed in the L<INPUT> section.

=head2 post

Post events to the object. First argument is the event to post, second is
the data hashref, following arguments are sent as arguments to the resultant
post.

  $poe_kernel->post( $alias => 'request',
                        open => { event => 'result' }, "localhost" );

=head2 yield

This method provides an alternative object based means of posting events to
the component. First argument is the event to post, second is the data
hashref, following arguments are sent as arguments to the resultant post.

  $poco->yield( open => { event => 'result' }, "localhost" );

=head2 call

This method provides an alternative object based means of calling events to
the component. First argument is the event to call, second is the data
hashref, following arguments are following arguments are
sent as arguments to the resultant call.

  $poco->call( open => { event => 'result' }, "localhost" );



=head2 simple method

All methods of the object can be called, but the first param must be the
data hashref as noted below in the L<INPUT> section below.

For example:
    
    $poco->open( { event => 'opened' },"localhost" );




=head1 INPUT

Each event requires a data hashref as an argument.

The data hashref may have the following keys.

=over 4

=item event

Event in your session that you want the results of the method
to go to.  C<event> is needed for all requests that you want a
response to.
No response is sent if C<event> is missing.

=item wantarray

Should the method be called in array context (1), scalar context (0) or
void context (undef)?  This will be set to a reasonable value for you
if you don't specify it.

=back

It is possible to pass arbitary data in the request hashref that could be
used in the resultant event handler. Simply define additional key/value
pairs of your own. It is recommended that one prefixes keys with '__' to
avoid future clashes.




=head1 OUTPUT

For each requested operation an event handler is required. C<ARG0> of this
event handler contains the data hashref.  C<ARG1..$#_> are the returned
values.


=over 4

=item state

The method that was called.

=item result

This is data returned from the function you called.  Usually a scalar, but can be
an array ref when using 'option_callback'.

=item error

In the event of an error occurring this will be defined. It is an scalar which
contains the error.

=back



=head1 STATUS

For your comfort and conveinence, the child process update C<$O> to
tell you what it is doing.  On many systems, C<$O> is available via
C</proc/$PID/cmdline>.


=head1 AUTHOR

Philip Gwyn E<lt>gwyn-at-cpan.orgE<gt>

Based on work by David Davis E<lt>xantus@cpan.orgE<gt>

=head1 SEE ALSO

L<POE>

=head1 RATING

Please rate this module. L<http://cpanratings.perl.org/rate/?distribution=POE-Component-Client-DAAP>

=head1 BUGS

Probably.  Report them here: L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=POE%3A%3AComponent%3A%3AClient%3A%3ADAAP>

=head1 CREDITS

BinGOs for L<POE::Component::Win32::Service> that helped me get started.

=head1 COPYRIGHT AND LICENSE

Copyright 2006 by Philip Gwyn;

Copyright 2005 by David Davis and Teknikill Software.

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

